'''pytests for testing a few different tools within opencos.eda'''

# pylint: disable=R0801 # (similar lines in 2+ files)

import os
import sys
import pytest

from opencos import eda, eda_tool_helper, eda_base

from opencos.tools.verilator import ToolVerilator
from opencos.tools.vivado import ToolVivado
from opencos.tools.cocotb import ToolCocotb
from opencos.tests import helpers
from opencos.tests.helpers import eda_wrap, eda_wrap_is_sim_fail
from opencos.utils.markup_helpers import yaml_safe_load
from opencos.utils import status_constants


thispath = os.path.dirname(__file__)

def chdir_remove_work_dir(relpath):
    '''Changes dir to relpath, removes the work directories (eda.work, eda.export*)'''
    return helpers.chdir_remove_work_dir(thispath, relpath)

# Figure out what tools the system has available, without calling eda.main(..)
config, tools_loaded = eda_tool_helper.get_config_and_tools_loaded()

def test_tools_loaded():
    '''Does not directly call 'eda.main' instead create a few Tool

    class objects and confirm the versioning methods work.
    '''
    assert config
    assert len(config.keys()) > 0

    # It's possible we're running in some container or install that has no tools, for example,
    # Windows.
    if sys.platform.startswith('win') and \
       not helpers.can_run_eda_command('elab', 'sim', config=config):
        # Windows, not handlers for elab or sim:
        pass
    else:
        assert len(tools_loaded) > 0

    def version_checker(
            obj: eda_base.Tool, chk_str: str
    ) -> None:
        assert obj.get_versions()
        full_ver = obj.get_full_tool_and_versions()
        assert chk_str in full_ver, f'{chk_str=} not in {full_ver=}'
        ver_num = full_ver.rsplit(':', maxsplit=1)[-1]
        if 'b' in ver_num:
            ver_num = ver_num.split('b')[0] # TODO(chaitanya): remove once cocotb 2.0 is released
        if '.' in ver_num:
            major_ver = ver_num.split('.')[0]
            assert major_ver.isdigit(), (
                f'Major version {major_ver=} is not a digit, from {full_ver=}'
            )
            assert float(major_ver) >= 0, (
                f'{major_ver=} is not a valid version number, from {full_ver=}'
            )
        else:
            assert float(ver_num), f'{ver_num=} is not a float, from {full_ver=}'


    # Do some very crude checks on the eda.Tool methods, and make
    # sure versions work for Verilator and Vivado:
    if 'verilator' in tools_loaded:
        my_tool = ToolVerilator(config={})
        version_checker(obj=my_tool, chk_str='verilator:')

    if 'vivado' in tools_loaded:
        my_tool = ToolVivado(config={})
        version_checker(obj=my_tool, chk_str='vivado:')

    if 'cocotb' in tools_loaded:
        my_tool = ToolCocotb(config={})
        version_checker(obj=my_tool, chk_str='cocotb:')

# Run these on simulation tools.
list_of_commands = [
    'sim',
    'elab'
]

list_of_tools = [
    'iverilog',
    'verilator',
    'vivado',
    'modelsim_ase',
    'questa_fse',
]

list_of_deps_targets = [
    ('tb_no_errs', True),       # target:str, sim_expect_pass:bool (sim only, all elab should pass)
    ('tb_dollar_fatal', False),
    ('tb_dollar_err', False),
]

list_of_added_sim_args = [
    '',
    '--gui --test-mode',
]

cannot_use_cocotb = 'cocotb' not in tools_loaded or \
    ('iverilog' not in tools_loaded and \
     'verilator' not in tools_loaded)
CANNOT_USE_COCOTB_REASON = 'requires cocotb in tools_loaded, and one of (iverilog, verilator) too'

@pytest.mark.parametrize("command", list_of_commands)
@pytest.mark.parametrize("tool", list_of_tools)
@pytest.mark.parametrize("target,sim_expect_pass", list_of_deps_targets)
@pytest.mark.parametrize("added_sim_args_str", list_of_added_sim_args)
def test_sim_elab_tools_pass_or_fail(command, tool, target, sim_expect_pass, added_sim_args_str):
    '''tests that: eda <sim|elab> --tool <parameter-tool> <parameter-args> <parameter-target>

    will correctly pass or fail depending on if it is supported or not.

    Also tests for: non-gui, or --gui --test-mode (runs non-gui, but most python args will
    be for --gui mode, signal logging, etc).
    '''
    if tool not in tools_loaded:
        pytest.skip(f"{tool=} skipped, {tools_loaded=}")
        return # skip/pass

    added_args = []
    if command == 'sim':
        added_args = added_sim_args_str.split()

    relative_dir = "deps_files/test_err_fatal"
    os.chdir(os.path.join(thispath, relative_dir))
    rc = eda.main(command, '--tool', tool, *(added_args), target)
    print(f'{rc=}')
    if command != 'sim' or sim_expect_pass:
        # command='elab' should pass.
        assert rc == 0
    else:
        assert eda_wrap_is_sim_fail(rc)


@pytest.mark.skipif('vivado' not in tools_loaded, reason="requires vivado")
def test_vivado_tool_defines():
    '''This test attempts to confirm that the following class inheritance works:

    Command <- CommandDesign <- CommandSim <- CommandSimVivado <- CommandElabVivado

    in particular that CommandElabVivado(CommandSimVivado, ToolVivado) has the
    correct ToolVivado.set_tool_defines() method, and that no other parent Command
    class has overriden it to defeat the defines that should be set.

    We also run with an added dependency (lib_ultrascale_plus_defines) to check that
    defines are set as expected.
    '''

    chdir_remove_work_dir('../../lib')
    rc = eda_wrap(
        'elab', '--tool', 'vivado', 'third_party/vendors/xilinx/lib_ultrascale_plus_defines',
        'oclib_fifo'
    )
    assert rc == 0

    # Confirm that args and defines we expected to be set are set.
    eda_config_yml_path = os.path.join(
        os.getcwd(), 'eda.work', 'oclib_fifo.elab', 'eda_output_config.yml'
    )

    data = yaml_safe_load(eda_config_yml_path)
    assert 'args' in data
    assert data['args'].get('top', '') == 'oclib_fifo'
    assert 'config' in data
    assert 'eda_original_args' in data['config']
    assert 'oclib_fifo' in data['config']['eda_original_args']
    assert data.get('target', '') == 'oclib_fifo'


    # This checks opencos.tools.vivado.ToolVivado.set_tool_defines():
    # We ran with --xilinx, so we expect certain defines to be set, others not to be set.
    assert 'defines' in data

    assert 'OC_TOOL_VIVADO' in data['defines']
    assert 'OC_LIBRARY' in data['defines']
    assert 'OC_LIBRARY_ULTRASCALE_PLUS' in data['defines']

    assert 'OC_LIBRARY_BEHAVIORAL' not in data['defines']
    assert 'VERILATOR' not in data['defines']
    assert 'SYNTHESIS' not in data['defines']

    assert data['defines']['OC_LIBRARY'] == '1'
    assert data['defines']['OC_LIBRARY_ULTRASCALE_PLUS'] is None # key present, no value


class TestCocotb:
    '''Namespace class for cocotb tests'''

    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_tool_defines(self):
        '''Test cocotb tool defines, configs, and integration.'''

        chdir_remove_work_dir('../../examples/cocotb')

        # Test 0: Using eda multi:
        rc = eda_wrap('multi', 'sim', '--tool=cocotb', '*test')
        assert rc == 0

        # Test 1: basic cocotb sim command with Python runner (default)
        rc = eda_wrap('sim', '--tool', 'cocotb', 'cocotb_counter_test')
        assert rc == 0

        # Test 2: cocotb works with different simulators/configurations
        rc = eda_wrap('sim', '--tool', 'cocotb', 'cocotb_counter_waves_test')
        assert rc == 0

        # Test 3: Makefile approach
        rc = eda_wrap('sim', '--tool', 'cocotb', 'cocotb_counter_makefile_test')
        assert rc == 0

        # Test 4: cocotb-specific defines are set correctly
        eda_config_yml_path = os.path.join(
            os.getcwd(), 'eda.work', 'cocotb_counter_test.sim', 'eda_output_config.yml'
        )

        data = yaml_safe_load(eda_config_yml_path)
        assert 'args' in data
        assert data['args'].get('top', '') == 'counter'
        assert 'config' in data
        assert 'eda_original_args' in data['config']
        assert 'cocotb_counter_test' in data['config']['eda_original_args'] or \
            './cocotb_counter_test' in data['config']['eda_original_args']
        assert data.get('target', '') == 'cocotb_counter_test'

        assert 'defines' in data
        assert 'OC_TOOL_COCOTB' in data['defines']
        assert 'SIMULATION' in data['defines']
        assert 'COCOTB' in data['defines']

        assert data['defines']['SIMULATION'] == 1
        assert data['defines']['COCOTB'] == 1
        assert data['defines']['OC_TOOL_COCOTB'] is None  # key present, no value

        assert 'VERILATOR' not in data['defines']
        assert 'SYNTHESIS' not in data['defines']



    @pytest.mark.parametrize("cocotb_simulator", ['verilator', 'iverilog'])
    @pytest.mark.parametrize("waves_arg", ["", "--waves"])
    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_different_simulators(self, cocotb_simulator, waves_arg):
        '''Test cocotb with different simulator configurations.'''

        if cocotb_simulator not in tools_loaded:
            pytest.skip(f"{cocotb_simulator=} skipped, {tools_loaded=}")
            return #skip/bypass

        chdir_remove_work_dir('../../examples/cocotb')

        rc = eda_wrap(
            'sim', '--tool', 'cocotb',
            f'--cocotb-simulator={cocotb_simulator}',
            waves_arg,
            'cocotb_counter_test',
        )
        assert rc == 0



    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_tool_instantiation(self):
        '''Test that ToolCocotb can be instantiated and has correct properties.'''

        tool = ToolCocotb(config={})

        # version detection works
        version = tool.get_versions()
        assert version, "Should return a non-empty version string"
        assert isinstance(version, str)

        # tool defines
        tool.set_tool_defines()
        defines = tool.defines
        assert 'SIMULATION' in defines
        assert 'COCOTB' in defines
        assert 'OC_TOOL_COCOTB' in defines
        assert defines['SIMULATION'] == 1
        assert defines['COCOTB'] == 1



    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_failure_cases(self):
        '''Test cocotb failure scenarios to ensure proper error handling.'''

        chdir_remove_work_dir('../../examples/cocotb')

        # Test 1: missing test files should fail gracefully
        rc = eda_wrap('sim', '--tool', 'cocotb', 'counter')  # Just HDL, no test files
        assert rc == status_constants.EDA_DEPS_TARGET_NOT_FOUND, \
            "Should fail when no cocotb test files are found"

        # Test 2: non-existent target should fail
        rc = eda_wrap('sim', '--tool', 'cocotb', 'nonexistent_target')
        assert rc in (
            status_constants.EDA_DEPS_TARGET_NOT_FOUND,
            # b/c we run eda_wrap, eda.main will continue to run after first error.
            status_constants.EDA_COMMAND_MISSING_TOP
        ), "Should fail for non-existent target"

        # Test 3: invalid cocotb test module should fail
        rc = eda_wrap(
            'sim', '--tool', 'cocotb',
            '--cocotb-test-module=nonexistent_test',
            'cocotb_counter_test',
        )
        assert eda_wrap_is_sim_fail(rc), \
            f"Should fail with invalid test module {rc=}"


    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_missing_dependencies(self):
        '''Test cocotb behavior when dependencies are missing.'''

        # Test missing cocotb installation (simulate by checking error handling)
        tool = ToolCocotb(config={})
        version = tool.get_versions()
        assert version, "Should return version when cocotb is properly installed"

        # Test tool defines are properly set even with minimal config
        tool.set_tool_defines()
        defines = tool.defines
        assert 'SIMULATION' in defines
        assert 'COCOTB' in defines
        assert 'OC_TOOL_COCOTB' in defines


    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_invalid_simulator(self):
        '''Test cocotb with invalid simulator configuration.'''

        chdir_remove_work_dir('../../examples/cocotb')

        rc = eda_wrap(
            'sim', '--tool', 'cocotb',
            '--cocotb-simulator=invalid_sim',
            'cocotb_counter_test',
        )
        assert eda_wrap_is_sim_fail(rc), \
            f"Should fail with invalid simulator {rc=}"

    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_malformed_hdl(self):
        '''Test cocotb with malformed HDL files.'''

        chdir_remove_work_dir('../../lib/tests')

        # Test with a target that has syntax errors - should fail during compilation
        rc = eda_wrap(
            'sim', '--tool', 'cocotb',
            '--cocotb-test-module=test_counter',
            'tb_dollar_fatal',
        )

        # eda_wrap may continue to errors beyond normal sim fails:
        assert eda_wrap_is_sim_fail(rc) or \
            f"Should fail with malformed HDL or failing test assertions {rc=}"


    @pytest.mark.skipif(cannot_use_cocotb, reason=CANNOT_USE_COCOTB_REASON)
    def test_cocotb_test_failures(self):
        '''Test that cocotb properly reports test failures.'''

        chdir_remove_work_dir('../../examples/cocotb')

        # Intentionally failing cocotb tests
        rc = eda_wrap('sim', '--tool', 'cocotb', 'cocotb_failure_test')
        assert eda_wrap_is_sim_fail(rc), \
            f"Should fail when cocotb tests contain assertion failures {rc=}"



''' opencos.tools.riviera - Used by opencos.eda for sim/elab commands w/ --tool=riviera.

Contains classes for ToolRiviera, CommandSimRiviera, CommandElabRiviera.
'''

# pylint: disable=too-many-ancestors
# pylint: disable=R0801 # (duplicate code in derived classes, such as if-condition return.)

import os
import shutil
import subprocess

from opencos import util
from opencos.tools.modelsim_ase import ToolModelsimAse, CommandSimModelsimAse
from opencos.utils.str_helpers import sanitize_defines_for_sh

class ToolRiviera(ToolModelsimAse):
    '''ToolRiviera used by opencos.eda for --tool=riviera'''

    _TOOL = 'riviera'
    _EXE = 'vsim'
    use_vopt = False

    def get_versions(self) -> str:
        if self._VERSION:
            return self._VERSION
        path = shutil.which(self._EXE)
        if not path:
            self.error(f"{self._EXE} not in path, need to setup or add to PATH")
            util.debug(f"{path=}")
        else:
            self.sim_exe = path
            self.sim_exe_base_path, _ = os.path.split(path)

        version_ret = subprocess.run(
            [self.sim_exe, '-version'],
            capture_output=True,
            check=False
        )
        stdout = version_ret.stdout.decode('utf-8', errors='replace').rstrip()

        # Expect:
        #  Aldec, Inc. Riviera-PRO version 2025.04.139.9738 built for Linux64 on May 30, 2025
        left, right = stdout.split('version')
        if 'Riviera' not in left:
            self.error(f'{stdout}: does not show Riviera')
        self._VERSION = right.split()[0]
        return self._VERSION


class CommandSimRiviera(CommandSimModelsimAse, ToolRiviera):
    '''CommandSimRiviera is a command handler for: eda sim --tool=riviera'''

    def __init__(self, config: dict):
        CommandSimModelsimAse.__init__(self, config=config)
        ToolRiviera.__init__(self, config=self.config)
        self.shell_command = os.path.join(self.sim_exe_base_path, 'vsim')
        self.starter_edition = True
        self.args.update({
            'tool': self._TOOL, # override
            'gui': False,
        })

    def set_tool_defines(self):
        # Update any defines from config.tools.modelsim_ase:
        self.defines.update(
            self.tool_config.get(
                'defines',
                # defaults, if not set:
                {'OC_TOOL_RIVIERA': 1}
            )
        )

    # Note: many of these we follow the same flow as CommandSimModelsimAse:
    # do_it, prepare_compile, compile, elaborate, simulate

    def compile(self):
        '''Override for CommandSimModelsimAse.compile() so we can set our own must_strings'''
        if self.args['stop-before-compile']:
            # don't run anything, save everyting we've already run in _prep_compile()
            return
        if self.args['stop-after-compile']:
            vsim_command_lists = self.get_compile_command_lists()
            self.run_commands_check_logs(vsim_command_lists, log_filename='sim.log',
                                         must_strings=['Compile success 0 Errors'],
                                         use_must_strings=False)

    def get_compile_command_lists(self, **kwargs) -> list:
        # This will also set up a compile.
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-l', 'sim.log', '-do', 'vsim_vlogonly.do'
        ]
        return [vsim_command_list]

    def get_elaborate_command_lists(self, **kwargs) -> list:
        # This will also set up a compile, for vlog + vsim (0 time)
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-l', 'sim.log', '-do', 'vsim_lintonly.do',
        ]
        return [vsim_command_list]

    def get_simulate_command_lists(self, **kwargs) -> list:
    # This will also set up a compile, for vlog + vsim (with run -all)
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-l', 'sim.log', '-do', 'vsim.do',
        ]
        return [vsim_command_list]

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []


    def write_vlog_dot_f(self, filename='vlog.f') -> None:
        '''Returns none, creates filename (str) for a vlog.f'''
        vlog_dot_f_lines = []

        # Add compile args from config.tool.riviera
        vlog_dot_f_lines += self.tool_config.get(
            'compile-args',
            '-sv -input_ports net').split()

        # Add waivers from config.tool.riviera, convert to warning:
        for waiver in self.tool_config.get('compile-waivers', []):
            vlog_dot_f_lines += [f'-err {waiver} W1']

        vlog_dot_f_fname = filename
        vlog_dot_f_fpath = os.path.join(self.args['work-dir'], vlog_dot_f_fname)

        for value in self.incdirs:
            vlog_dot_f_lines += [ f"+incdir+{value}" ]

        for k,v in self.defines.items():
            if v is None:
                vlog_dot_f_lines += [ f'+define+{k}' ]
            else:
                # if the value v is a double-quoted string, such as v='"hi"', the
                # entire +define+NAME="hi" needs to wrapped in double quotes with the
                # value v double-quotes escaped: "+define+NAME=\"hi\""
                if isinstance(v, str) and v.startswith('"') and v.endswith('"'):
                    str_v = v.replace('"', '\\"')
                    vlog_dot_f_lines += [ f'"+define+{k}={str_v}"' ]
                else:
                    # Generally we should only support int and str python types passed as
                    # +define+{k}={v}, but also for SystemVerilog plusargs
                    vlog_dot_f_lines += [ f'+define+{k}={sanitize_defines_for_sh(v)}' ]


        vlog_dot_f_lines += self.args['compile-args']
        if self.args['coverage']:
            vlog_dot_f_lines += self.tool_config.get('compile-coverage-args', '').split()

        vlog_dot_f_lines += [
            ] + list(self.files_sv) + list(self.files_v)

        if not self.files_sv and not self.files_v:
            if not self.args['stop-before-compile']:
                self.error(f'{self.target=} {self.files_sv=} and {self.files_v=} are empty,',
                           'cannot create a valid vlog.f')

        with open(vlog_dot_f_fpath, 'w', encoding='utf-8') as f:
            f.writelines(line + "\n" for line in vlog_dot_f_lines)


    def write_vsim_dot_do(self, dot_do_to_write: list) -> None:
        '''Writes files(s) based on dot_do_to_write(list of str)

        list arg values can be empty (all) or have items 'all', 'sim', 'lint', 'vlog'.'''

        vsim_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim.do')
        vsim_lintonly_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim_lintonly.do')
        vsim_vlogonly_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim_vlogonly.do')

        sim_plusargs_str = self._get_sim_plusargs_str()

        voptargs_str = ""
        if self.args['gui'] or self.args['waves'] or self.args['coverage']:
            voptargs_str += self.tool_config.get('simulate-waves-args',
                                                 '+accb +accr +access +r+w')
        if self.args['coverage']:
            voptargs_str += self.tool_config.get('coverage-args', '')

        # parameters
        if self.parameters:
            voptargs_str += ' ' + ' '.join(self.process_parameters_get_list(arg_prefix='-G'))

        # TODO(drew): support self.args['sim_libary', 'elab-args', sim-args'] (3 lists)
        # to add to vsim_one_liner.

        vsim_one_liner = (
            "vsim"
            f" -sv_seed {self.args['seed']} {sim_plusargs_str}"
            f" {voptargs_str} work.{self.args['top']}"
        )

        vsim_one_liner = vsim_one_liner.replace('\n', ' ') # needs to be a one-liner

        vsim_vlogonly_dot_do_lines = [
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -code 20 -force;",
            "    }",
            "}",
            "if {[batch_mode]} {",
            "    quit -code 0 -force;",
            "}",
        ]

        vsim_lintonly_dot_do_lines = [
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "set qc 30;",
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -code 20 -force;",
            "    }",
            "}",
            "if {[catch { " + vsim_one_liner + " } result] } {",
            "    echo \"Caught $result\";",
            "    if {[batch_mode]} {",
            "        quit -code 19 -force;",
            "    }",
            "}",
            "if {[batch_mode]} {",
            "    quit -code 0 -force;",
            "}",
        ]

        vsim_dot_do_lines = [
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "set qc 30;",
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -code 20 -force;",
            "    }",
            "}",
            "if {[catch { " + vsim_one_liner + " } result] } {",
            "    echo \"Caught $result\";",
            "    if {[batch_mode]} {",
            "        quit -code 19 -force;",
            "    }",
            "}",
        ]

        if self.args['coverage']:
            vsim_dot_do_lines += [
                "run -all;",
                "acdb save",
                "acdb report -db work.acdb -txt -o cov.txt",
                # Note - could try:
                ##"cover report -o cov.report.txt -fullverbose -all_columns",
            ]
        else:
            vsim_dot_do_lines += [
                "run -all;",
            ]


        vsim_dot_do_lines += [
            "if {[batch_mode]} {",
            "    quit -code 0 -force;",
            "}",
        ]

        write_all = len(dot_do_to_write) == 0 or 'all' in dot_do_to_write
        if write_all or 'sim' in dot_do_to_write:
            with open(vsim_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_dot_do_lines)

        if write_all or 'lint' in dot_do_to_write:
            with open(vsim_lintonly_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_lintonly_dot_do_lines)

        if write_all or 'vlog' in dot_do_to_write:
            with open(vsim_vlogonly_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_vlogonly_dot_do_lines)



    def _get_vsim_suppress_list_str(self) -> str:
        vsim_suppress_list = []
        # Add waivers from config.tool.modelsim_ase:
        for waiver in self.tool_config.get(
                'simulate-waivers', [
                    #defaults: none
                ]):
            vsim_suppress_list += ['-filter', str(waiver)]

        return ' '.join(vsim_suppress_list)


class CommandElabRiviera(CommandSimRiviera):
    '''CommandElabRiviera is a command handler for: eda elab --tool=riviera'''

    command_name = 'elab'

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-elaborate'] = True


class CommandLintRiviera(CommandSimRiviera):
    '''CommandLintRiviera is a command handler for: eda lint --tool=riviera'''

    command_name = 'lint'

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-compile'] = True
        self.args['stop-after-elaborate'] = True



'''pytests for: eda [multi|tools-multi] synth [args] <target(s)>'''

import os
import shutil

import pytest

from opencos import eda, eda_tool_helper
from opencos.tests import helpers
from opencos.tests.helpers import Helpers
from opencos.utils.markup_helpers import yaml_safe_load


THISPATH = os.path.dirname(__file__)

def chdir_remove_work_dir(relpath):
    '''Changes dir to relpath, removes the work directories (eda.work, eda.export*)'''
    return helpers.chdir_remove_work_dir(THISPATH, relpath)

# Figure out what tools the system has available, without calling eda.main(..)
config, tools_loaded = eda_tool_helper.get_config_and_tools_loaded()

# list of tools we'd like to try:
list_of_synth_tools = [
    'invio_yosys',
    'tabbycad_yosys',
    'slang_yosys'
]

list_of_elab_tools = [
    'invio_yosys',
    'slang_yosys'
]

def skip_it(tool, command) -> bool:
    '''skip_it: returns True if we should skip due to lack of tool existence'''
    return bool( tool not in tools_loaded or
                 (command == 'elab' and tool not in list_of_elab_tools) or
                 (command == 'synth' and tool not in list_of_synth_tools) )


@pytest.mark.parametrize("tool", list_of_synth_tools)
@pytest.mark.parametrize("command", ['elab', 'synth'])
class Tests:
    '''skippable (via pytest parameters) class holder for pytest methods'''

    def test_args_multi_synth_bad_target_should_fail(self, tool, command):
        '''Tests: eda mulit <elab|synth>, and this test should fail.'''
        if skip_it(tool, command):
            pytest.skip(f"{tool=} {command=} skipped, {tools_loaded=}")
            return # skip/pass
        chdir_remove_work_dir('../../lib')
        cmdlist = (f'multi {command} --debug --fail-if-no-targets --tool {tool}'
                   ' target_doesnt_exist*').split()
        rc = eda.main(*cmdlist)
        print(f'{rc=}')
        assert rc > 1

    def test_args_multi_synth_oclib_fifos(self, tool, command):
        '''This should be 4 jobs and takes ~15 seconds for synth.'''
        if skip_it(tool, command):
            pytest.skip(f"{tool=} {command=} skipped, {tools_loaded=}")
            return # skip/pass
        chdir_remove_work_dir('../../lib')
        cmdlist = (f'multi {command} --debug --fail-if-no-targets --tool {tool}'
                   ' --yosys-synth=synth_xilinx oclib_fifo*').split()
        rc = eda.main(*cmdlist)
        print(f'{rc=}')
        assert rc == 0

    def test_args_multi_synth_oclib_rams(self, tool, command):
        '''This should be 4 jobs and takes ~15 seconds for synth.'''
        if skip_it(tool, command):
            pytest.skip(f"{tool=} {command=} skipped, {tools_loaded=}")
            return # skip/pass
        chdir_remove_work_dir('../../lib')
        cmdlist = (f'multi {command} --debug --fail-if-no-targets --tool {tool}'
                   ' --yosys-synth=synth_xilinx rams/oclib_ram*').split()
        rc = eda.main(*cmdlist)
        print(f'{rc=}')
        assert rc == 0


def vivado_has_xpms() -> bool:
    '''Returns True if Vivado is installed and has visibility to XPMs'''
    if 'vivado' not in tools_loaded:
        return False
    vivado_exe = shutil.which('vivado')
    vivado_bin_path, _ = os.path.split(vivado_exe)
    vivado_base_path, _ = os.path.split(vivado_bin_path) # strip bin/vivado

    return os.path.exists(os.path.join(vivado_base_path, 'data', 'ip', 'xpm'))


@pytest.mark.skipif(
    'slang_yosys' not in tools_loaded, reason="requires slang_yosys for synth"
)
class TestsSlangYosys(Helpers):
    '''Tests that require tool=slang_yosys to be available'''

    def test_sdc_file(self):
        '''Test for 'eda synth' on oclib_fifo_with_sdc

        This does not use the actual .sdc file, but it also shouldn't fail with
        that file in the 'deps' list (CommandDesign should track it correctly)
        '''
        chdir_remove_work_dir('deps_files/test_sdc_files')
        cmd_str = 'synth --tool=slang_yosys oclib_fifo_with_sdc'
        rc = self.log_it(cmd_str, use_eda_wrap=False)
        assert rc == 0

        # Since vanilla yosys won't use the SDC file, let's at least confirm
        # that EDA used it and tracked it:
        eda_config_yml_path = os.path.join(
            os.getcwd(), 'eda.work', 'oclib_fifo_with_sdc.synth', 'eda_output_config.yml'
        )
        data = yaml_safe_load(eda_config_yml_path)
        assert 'files_sdc' in data
        assert data['files_sdc']
        assert data['files_sdc'][0].endswith('oclib_fifo_yosys.sdc')


@pytest.mark.skipif('vivado' not in tools_loaded, reason="requires vivado")
@pytest.mark.skipif(not vivado_has_xpms(), reason="requires install to have XPMs")
class TestsVivado(Helpers):
    '''Tests that require tool=vivado with XPMs available for synthesis'''

    def test_sdc_file(self):
        '''Test for 'eda synth' on oclib_fifo_with_sdc

        And check that the .sdc file was used and not the default generated .xdc
        file
        '''
        chdir_remove_work_dir('deps_files/test_sdc_files')
        cmd_str = 'synth --tool=vivado oclib_fifo_with_sdc'
        rc = self.log_it(cmd_str, use_eda_wrap=False)
        assert rc == 0

        # apparently this doesn't get saved in the eda.log. That's not great,
        # but we can check the <target>.synth.log file.
        synth_log = os.path.join(
            THISPATH, 'deps_files', 'test_sdc_files',
            'eda.work', 'oclib_fifo_with_sdc.synth', 'oclib_fifo.synth.log'
        )
        sdc_lines = self.get_log_lines_with('.sdc', logfile=synth_log)
        assert sdc_lines
        for sdc_line in sdc_lines:
            assert 'oclib_fifo_vivado.sdc' in sdc_line
            assert 'test_sdc_files' in sdc_line
            assert 'eda.work' not in sdc_line



''' opencos.tools.questa - Used by opencos.eda commands with --tool=questa

Also a base class for tools.modelsim_ase.
'''

# pylint: disable=R0801 # (setting similar, but not identical, self.defines key/value pairs)

# TODO(drew): fix these pylint eventually:
# pylint: disable=too-many-branches

import os
import re
import shutil

from opencos import util
from opencos.eda_base import Tool
from opencos.commands import CommandSim, CommandFList

class ToolQuesta(Tool):
    '''Base class for CommandSimQuesta, collects version information about qrun'''

    _TOOL = 'questa'
    _EXE = 'qrun'

    starter_edition = False # Aka, modelsim_ase
    use_vopt = shutil.which('vopt') # vopt exists in qrun/vsim framework, and we'll use it.
    sim_exe = '' # vsim or qrun
    sim_exe_base_path = ''
    questa_major = None
    questa_minor = None

    def __init__(self, config: dict):
        super().__init__(config=config)
        self.args['part'] = 'xcu200-fsgd2104-2-e'

    def get_versions(self) -> str:
        if self._VERSION:
            return self._VERSION
        path = shutil.which(self._EXE)
        if not path:
            self.error(f"{self._EXE} not in path, need to setup",
                       "(i.e. source /opt/intelFPGA_pro/23.4/settings64.sh")
            util.debug(f"{path=}")
            if self._EXE.endswith('qrun') and \
               any(x in path for x in ('modelsim_ase', 'questa_fse')):
                util.warning(f"{self._EXE=} Questa path is for starter edition",
                             "(modelsim_ase, questa_fse), consider using --tool=modelsim_ase",
                             "or --tool=questa_fse")
        else:
            self.sim_exe = path
            self.sim_exe_base_path, _ = os.path.split(path)

        if self._EXE.endswith('vsim'):
            self.starter_edition = True

        m = re.search(r'(\d+)\.(\d+)', path)
        if m:
            self.questa_major = int(m.group(1))
            self.questa_minor = int(m.group(2))
            self._VERSION = str(self.questa_major) + '.' + str(self.questa_minor)
        else:
            self.error("Questa path doesn't specificy version, expecting (d+.d+)")
        return self._VERSION

    def set_tool_defines(self):
        # Will only be called from an object which also inherits from CommandDesign,
        # i.e. has self.defines
        self.defines['OC_TOOL_QUESTA'] = None
        self.defines[f'OC_TOOL_QUESTA_{self.questa_major:d}_{self.questa_minor:d}'] = None

class CommandSimQuesta(CommandSim, ToolQuesta):
    '''Command handler for: eda sim --tool=questa.'''

    def __init__(self, config:dict):
        CommandSim.__init__(self, config)
        ToolQuesta.__init__(self, config=self.config)
        # add args specific to this simulator
        self.args.update({
            'gui': False,
            'tcl-file': 'sim.tcl',
            'work-lib': 'work',
        })
        self.args_help.update({
            'gui': 'Run Questa in GUI mode',
            'tcl-file': 'name of TCL file to be created for Questa simulation',
            'work-lib': 'Questa work library name',
        })

        self.shell_command = self.sim_exe # set by ToolQuesta.get_versions(self)
        self.vlog_commands = []
        self.vopt_commands = []
        self.vsim_commands = []

    def set_tool_defines(self):
        ToolQuesta.set_tool_defines(self)

    # We do not override CommandSim.do_it(), CommandSim.check_logs_for_errors(...)

    def prepare_compile(self):
        self.set_tool_defines()
        self.vlog_commands = self.get_compile_command_lists()
        self.vopt_commands = self.get_elaborate_command_lists()
        self.vsim_commands = self.get_simulate_command_lists()
        self.write_sh_scripts_to_work_dir(
            compile_lists=self.vlog_commands,
            elaborate_lists=self.vopt_commands,
            simulate_lists=self.vsim_commands
        )

    def compile(self):
        if self.args['stop-before-compile']:
            return
        self.run_commands_check_logs(
            self.vlog_commands, check_logs=True, log_filename='vlog.log'
        )

    def elaborate(self):
        if self.args['stop-before-compile'] or self.args['stop-after-compile']:
            return
        self.run_commands_check_logs(
            self.vopt_commands, check_logs=True, log_filename='vopt.log'
        )

    def simulate(self):
        if self.args['stop-before-compile'] or self.args['stop-after-compile'] or \
           self.args['stop-after-elaborate']:
            return
        self.run_commands_check_logs(
            self.vsim_commands, check_logs=True, log_filename='vsim.log'
        )

    def get_compile_command_lists(self, **kwargs) -> list:
        self.set_tool_defines()
        ret = []

        # Create work library
        vlib_cmd = [os.path.join(self.sim_exe_base_path, 'vlib'), self.args['work-lib']]
        ret.append(vlib_cmd)

        # Map work library
        vmap_cmd = [os.path.join(self.sim_exe_base_path, 'vmap'), 'work', self.args['work-lib']]
        ret.append(vmap_cmd)

        # Compile files
        if self.files_v or self.files_sv:
            vlog_cmd = [os.path.join(self.sim_exe_base_path, 'vlog'), '-64', '-sv']

            # Add include directories
            for incdir in self.incdirs:
                vlog_cmd += [f'+incdir+{incdir}']

            # Add defines
            for key, value in self.defines.items():
                if value is None:
                    vlog_cmd += [f'+define+{key}']
                else:
                    vlog_cmd += [f'+define+{key}={value}']

            # Add suppression flags
            vlog_cmd += [
                '-svinputport=net',
                '-suppress', 'vlog-2275',
                '-suppress', 'vlog-2583',
            ]

            # Add source files
            vlog_cmd += self.files_v + self.files_sv

            ret.append(vlog_cmd)

        # Compile VHDL files if any
        if self.files_vhd:
            vcom_cmd = [os.path.join(self.sim_exe_base_path, 'vcom'), '-64']
            vcom_cmd += self.files_vhd
            ret.append(vcom_cmd)

        return ret

    def get_elaborate_command_lists(self, **kwargs) -> list:
        if self.args['stop-after-compile']:
            return []

        vopt_cmd = [os.path.join(self.sim_exe_base_path, 'vopt'), '-64']

        # Add optimization flags
        vopt_cmd += [
            '-suppress', 'vopt-13159',
            '-suppress', 'vopt-2685',
            '-note', 'vopt-2718',
        ]

        if self.args['gui'] or self.args['waves']:
            vopt_cmd += ['+acc']

        # Add top module and output
        vopt_cmd += [self.args['top'], '-o', 'opt_design']

        return [vopt_cmd]

    def get_simulate_command_lists(self, **kwargs) -> list:
        # Create TCL file
        tcl_name = os.path.abspath(os.path.join(self.args['work-dir'], self.args['tcl-file']))

        if self.args['waves']:
            util.artifacts.add_extension(
                search_paths=self.args['work-dir'], file_extension='wlf',
                typ='waveform', description='Questa Waveform WLF file'
            )

        with open(tcl_name, 'w', encoding='utf-8') as fo:
            if self.args['waves']:
                if self.args['waves-start']:
                    print(f"run {self.args['waves-start']} ns", file=fo)
                print("add wave -r /*", file=fo)
            print("run -all", file=fo)
            if self.run_in_batch_mode():
                print("quit", file=fo)

        # Create vsim command
        vsim_cmd = [os.path.join(self.sim_exe_base_path, 'vsim'), '-64']

        if not self.run_in_batch_mode():
            vsim_cmd += ['-gui']
        else:
            vsim_cmd += ['-c']

        if util.args['verbose']:
            vsim_cmd += ['-verbose']

        # Add simulation arguments
        vsim_cmd += ['-do', tcl_name, 'opt_design']

        return [vsim_cmd]

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []

    def run_in_batch_mode(self) -> bool:
        '''Returns bool if we should run in batch mode (-c) from command line'''
        if self.args['test-mode']:
            return True
        if self.args['gui']:
            return False
        return True

    def artifacts_add(self, name: str, typ: str, description: str) -> None:
        '''Override from Command.artifacts_add for better descriptions'''
        _, leafname = os.path.split(name)
        if leafname == 'vsim.log':
            description = 'Questa simulation step (3/3) log from stdout/stderr'
        elif leafname == 'vopt.log':
            description = 'Questa elaboration step (2/3) log from stdout/stderr'
        elif leafname == 'vlog.log':
            description = 'Questa compile step (1/3) log from stdout/stderr'

        super().artifacts_add(name=name, typ=typ, description=description)


class CommandFListQuesta(CommandFList, ToolQuesta):
    '''CommandFListQuesta is a command handler for: eda flist --tool=questa'''

    def __init__(self, config: dict):
        CommandFList.__init__(self, config=config)
        ToolQuesta.__init__(self, config=self.config)


class CommandElabQuesta(CommandSimQuesta):
    '''Command handler for: eda elab --tool=questa'''

    command_name = 'elab'

    def __init__(self, config:dict):
        CommandSimQuesta.__init__(self, config)
        # add args specific to this simulator
        self.args['stop-after-elaborate'] = True


class CommandLintQuesta(CommandSimQuesta):
    '''Command handler for: eda lint --tool=questa'''

    command_name = 'lint'

    def __init__(self, config:dict):
        CommandSimQuesta.__init__(self, config)
        # add args specific to this simulator
        self.args['stop-after-compile'] = True
        self.args['stop-after-elaborate'] = True



''' opencos.utils.subprocess_helpers -- wrappers for subprocess to support background/tee'''

import os
import shutil
import subprocess
import sys

from opencos.util import debug, error, info, progname, global_log

IS_WINDOWS = sys.platform.startswith('win')

def subprocess_run(
        work_dir: str, command_list: list, fake: bool = False, shell: bool = False
) -> int:
    ''' Run command_list in the foreground, with preference to use bash if shell=True'''

    proc_kwargs = {
        'shell': shell
    }
    if work_dir:
        proc_kwargs['cwd'] = work_dir

    bash_exec = shutil.which('bash')
    if shell and bash_exec and not IS_WINDOWS:
        proc_kwargs.update({'executable': bash_exec})

    if not IS_WINDOWS and shell:
        c = ' '.join(command_list)
    else:
        c = command_list

    if fake:
        info(f"subprocess_run FAKE: would have called subprocess.run({c}, **{proc_kwargs}")
        return 0

    debug(f"subprocess_run: About to call subprocess.run({c}, **{proc_kwargs}")
    proc = subprocess.run(c, check=True, **proc_kwargs)
    return proc.returncode


def subprocess_run_background(
        work_dir: str, command_list: list, background: bool = True, fake : bool = False,
        shell: bool = False, tee_fpath: str = ''
) -> (str, str, int):
    ''' Run command_list in the background, with preference to use bash if shell=True

    tee_fpath is relative to work_dir.

    Note that stderr is converted to stdout, and stderr is retuned as '':
        Returns tuple of (stdout str, '', int return code)
    '''

    debug(f'subprocess_run_background: {background=} {tee_fpath=} {shell=}')

    if fake:
        # let subprocess_run handle it (won't run anything)
        rc = subprocess_run(work_dir, command_list, fake=fake, shell=shell)
        return '', '', rc

    proc_kwargs = {'shell': shell,
                   'stdout': subprocess.PIPE,
                   'stderr': subprocess.STDOUT,
                   }
    if work_dir:
        proc_kwargs['cwd'] = work_dir

    bash_exec = shutil.which('bash')
    if shell and bash_exec and not IS_WINDOWS:
        # Note - windows powershell will end up calling: /bin/bash /c, which won't work
        proc_kwargs.update({'executable': bash_exec})

    if not IS_WINDOWS and shell:
        c = ' '.join(command_list)
    else:
        c = command_list # leave as list.

    debug(f"subprocess_run_background: about to call subprocess.Popen({c}, **{proc_kwargs})")
    proc = subprocess.Popen(c, **proc_kwargs) # pylint: disable=consider-using-with

    stdout = ''
    tee_fpath_f = None
    if tee_fpath:
        tee_fpath = os.path.join(work_dir, tee_fpath)
        try:
            tee_fpath_f = open( # pylint: disable=consider-using-with
                tee_fpath, 'w', encoding='utf-8'
            )
        except Exception as e:
            error(f'Unable to open file "{tee_fpath}" for writing, {e}')

    for line in iter(proc.stdout.readline, b''):
        line = line.rstrip().decode("utf-8", errors="replace")
        if not background:
            print(line)
        if tee_fpath_f:
            tee_fpath_f.write(line + '\n')
        if global_log.file:
            global_log.write(line, '\n')
        stdout += line + '\n'

    proc.communicate()
    rc = proc.returncode
    if tee_fpath_f:
        tee_fpath_f.write(f'INFO: [{progname}] subprocess_run_background: returncode={rc}\n')
        tee_fpath_f.close()
        info('subprocess_run_background: wrote: ' + os.path.abspath(tee_fpath))

    return stdout, '', rc



'''Because so many tools use the exe vsim, I don't want to run `vsim -version` N times for N
tools each to figure out if it's full-Questa, Modelsim, QuestaFSE, or Riviera.

Instead, eda.py can call this once, and then query if this tool exists when running
opencos.eda.auto_tool_setup(..)
'''

import shutil
import subprocess

from opencos.util import debug

vsim_path = shutil.which('vsim')

INIT_HAS_RUN = False
TOOL_IS = {
    'riviera': False,
    'modelsim_ase': False,
    'questa' : False,
    'questa_fse': False
}


def init() -> None:
    '''Sets INIT_HAS_RUN=True (only runs once) and one of TOOL_IS[tool] = True'''
    global INIT_HAS_RUN # pylint: disable=global-statement

    if INIT_HAS_RUN:
        return

    INIT_HAS_RUN = True

    if not vsim_path:
        return

    proc = None
    try:
        proc = subprocess.run([vsim_path, '-version'], capture_output=True, check=False)
    except Exception as e:
        debug(f'vsim -version: exception {e}')

    if proc is None or proc.returncode != 0:
        return


    stdout_str_lower = proc.stdout.decode('utf-8', errors='replace').lower()

    if all(x in stdout_str_lower for x in ('starter', 'modelsim', 'fpga')):
        TOOL_IS['modelsim_ase'] = True
    elif all(x in stdout_str_lower for x in ('starter', 'questa', 'fpga')):
        TOOL_IS['questa_fse'] = True
    elif all(x in stdout_str_lower for x in ('riviera', 'aldec')):
        TOOL_IS['riviera'] = True
    elif 'questa' in stdout_str_lower:
        TOOL_IS['questa'] = True



''' opencos.eda_config - handles --config-yml arg, and providing a
'config' dict for use by opencos.eda, opencos.commands, and opencos.tools

Order of precedence for default value of eda arg: --config-yaml:
1) os.environ.get('EDA_CONFIG_YML', '') -- filepath to an eda_config.yml file
2) ~/.opencos-eda/EDA_CONFIG.yml
3) (package pip installed dir for opencos)/eda_config_defaults.yml
'''

import copy
import os
import argparse
import shutil

import mergedeep

from opencos import util
from opencos.utils.markup_helpers import yaml_safe_load, yaml_safe_writer

class Defaults:
    '''Defaults is a global placeholder for constants and supported features.'''

    environ_override_config_yml = os.environ.get('EDA_CONFIG_YML', '')
    home_override_config_yml = os.path.join(
        os.environ.get('HOME', ''), '.opencos-eda', 'EDA_CONFIG.yml'
    )
    opencos_config_yml = 'eda_config_defaults.yml'
    config_yml = ''

    supported_config_keys = set([
        'DEFAULT_HANDLERS', 'DEFAULT_HANDLERS_HELP',
        'defines',
        'dep_command_enables',
        'dep_tags_enables',
        'deps_markup_supported',
        'deps_subprocess_shell',
        'bare_plusarg_supported',
        'dep_sub',
        'vars',
        'file_extensions',
        'command_determines_tool',
        'command_tool_is_optional',
        'tools',
        'auto_tools_order',
    ])
    supported_config_auto_tools_order_keys = set([
        'exe', 'handlers',
        'requires_env', 'requires_py', 'requires_cmd', 'requires_in_exe_path',
        'requires_vsim_helper', 'requires_vscode_extension',
        'disable-tools-multi',
    ])
    supported_config_tool_keys = set([
        'defines',
        'log-bad-strings',
        'log-must-strings',
        'sim-libraries',
        'compile-args',
        'compile-waves-args',
        'compile-waivers',
        'compile-coverage-args',
        'elab-args',
        'elab-waves-args',
        'simulate-args',
        'simulate-waves-args',
        'simulate-waivers',
        'coverage-args',
    ])

EDA_OUTPUT_CONFIG_FNAME = 'eda_output_config.yml'

if os.path.exists(Defaults.environ_override_config_yml):
    Defaults.config_yml = Defaults.environ_override_config_yml
elif os.path.exists(Defaults.home_override_config_yml):
    Defaults.config_yml = Defaults.home_override_config_yml
else:
    Defaults.config_yml = Defaults.opencos_config_yml


def find_eda_config_yml_fpath(
        filename:str, package_search_only=False, package_search_enabled=True
) -> str:
    '''Locates the filename (.yml) either from fullpath provided or from the sys.path
    (pip-installed) opencos-eda package paths.'''

    # Check fullpath, unless we're only checking the installed pacakge dir.
    if package_search_only:
        pass
    elif os.path.exists(filename):
        return os.path.abspath(filename)

    leaf_filename = os.path.split(filename)[1]

    if leaf_filename != filename:
        # filename had subdirs, and we didn't find it already.
        util.error(f'eda_config: Could not find {filename=}')
        return None

    # Search in . or pacakge installed dir
    thispath = os.path.dirname(__file__) # this is not an executable, should be in packages dir.

    if package_search_only:
        paths = [thispath]
    elif package_search_enabled:
        paths = ['', thispath]
    else:
        paths = ['']


    for dpath in paths:
        fpath = os.path.join(dpath, leaf_filename)
        if os.path.exists(fpath):
            return fpath

    util.error(f'eda_config: Could not find {leaf_filename=} in opencos within {paths=}')
    return None


def check_config(config:dict, filename='') -> None:
    '''Returns None, will util.error(..) if there are issues in 'config'

    checks for known dict keys and data types, is NOT exhaustive checking.
    '''

    # sanity checks:
    for key in config:
        if key not in Defaults.supported_config_keys:
            util.error(f'eda_config.get_config({filename=}): has unsupported {key=}' \
                       + f' {Defaults.supported_config_keys=}')

    for row in config.get('auto_tools_order', []):
        for tool, table in row.items():
            for key in table:
                if key not in Defaults.supported_config_auto_tools_order_keys:
                    util.error(f'eda_config.get_config({filename=}): has unsupported {key=}' \
                               + f' in auto_tools_order, {tool=},' \
                               + f' {Defaults.supported_config_auto_tools_order_keys=}')

    for tool,table in config.get('tools', {}).items():
        for key in table:
            if key not in Defaults.supported_config_tool_keys:
                util.error(f'eda_config.get_config({filename=}): has unsupported {key=}' \
                           + f' in config.tools.{tool=}, ' \
                           + f' {Defaults.supported_config_tool_keys=}')


def update_config_auto_tool_order_for_tool(tool: str, config: dict) -> str:
    '''Update config entry if the value for tool is in the form 'name=/path/to/exe

    Input arg tool can be in the form (for example):
      tool='verlator', tool='verilator=/path/to/verilator.exe'

    Performs no update if tool has no = in it. Returns tool (str) w/out = in it
    '''
    if not tool or '=' not in tool:
        return tool

    tool, user_exe = tool.split('=')[0:2]

    user_exe = shutil.which(user_exe)

    if tool not in config['auto_tools_order'][0]:
        return tool
    if not user_exe:
        return tool

    old_exe = config['auto_tools_order'][0][tool].get('exe', str())
    if isinstance(old_exe, list):
        config['auto_tools_order'][0][tool]['exe'][0] = user_exe
    else:
        config['auto_tools_order'][0][tool]['exe'] = user_exe
    return tool


def update_config_auto_tool_order_for_tools(tools: list, config: dict) -> list:
    '''Given a list of tools and eda_config style 'config' dict, update

    the auto_tool_order (consumed by opencos.eda when --tool is not specified).
    '''
    ret = []
    for tool in tools:
        ret.append(update_config_auto_tool_order_for_tool(tool, config))
    return ret


def update_config_for_eda_safe(config) -> None:
    '''Set method to update config dict values to run in a "safe" mode'''
    config['dep_command_enables']['shell'] = False


def deps_shell_commands_enabled(config) -> bool:
    '''Get method on config to determine if DEPS.yml shell-style commands are allowed'''
    return config['dep_command_enables']['shell']


def get_config(filename) -> dict:
    '''Given an eda_config_default.yml (or --config-yml=<filename>) return a config

    dict from the filename.'''

    fpath = find_eda_config_yml_fpath(filename)
    user_config = yaml_safe_load(fpath)
    check_config(user_config, filename=filename)

    # The final thing we do is update key 'config-yml' with the full path used.
    # This way we don't have to pass around --config-yml as some special arg
    # in eda.CommandDesign.args, and eda.CommandMulti can use when re-invoking 'eda'.
    user_config['config-yml'] = fpath
    return user_config


def get_config_handle_defaults(filename) -> dict:
    '''Given a user provided --config-yml=<filename>, return a merged config with

    the existing default config.'''

    user_config = get_config(filename)
    user_config = get_config_merged_with_defaults(user_config)
    return user_config


def merge_config(dst_config:dict, overrides_config:dict, additive_strategy=False) -> None:
    '''Mutates dst_config, uses Strategy.TYPESAFE_REPLACE'''
    # TODO(drew): It would be cool if I could have Sets be additive, but oh well,
    # this gives the user more control over replacing entire lists.
    strategy = mergedeep.Strategy.TYPESAFE_REPLACE
    if additive_strategy:
        strategy = mergedeep.Strategy.TYPESAFE_ADDITIVE
    mergedeep.merge(dst_config, overrides_config, strategy=strategy)


def get_config_merged_with_defaults(config:dict) -> dict:
    '''Returns a new config that has been merged with the default config.

    The default config location is based on Defaults.config_yml (env, local, or pip
    installed location)'''

    default_fpath = find_eda_config_yml_fpath(Defaults.config_yml, package_search_only=True)
    default_config = yaml_safe_load(default_fpath)
    merge_config(default_config, overrides_config=config)
    # This technically mutated updated into default_config, so return that one:
    return default_config


def get_argparser() -> argparse.ArgumentParser:
    '''Returns an ArgumentParser, handles --config-yml=<filename> arg'''
    parser = argparse.ArgumentParser(
        prog='opencos eda config options', add_help=False, allow_abbrev=False
    )
    parser.add_argument('--config-yml', type=str, default=Defaults.config_yml,
                        help=('YAML filename to use for configuration (default'
                              f' {Defaults.config_yml})'))
    return parser


def get_argparser_short_help() -> str:
    '''Returns a shortened help string given for arg --config-yml.'''
    return util.get_argparser_short_help(parser=get_argparser())


def get_eda_config(args:list, quiet=False) -> (dict, list):
    '''Returns an config dict and a list of args to be passed downstream
    to eda.main and eda.process_tokens.

    Handles args for:
      --config-yml=<YAMLFILE>

    This will merge the result with the default config (if overriden)
    '''

    parser = get_argparser()
    try:
        parsed, unparsed = parser.parse_known_args(args + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        util.error(f'problem attempting to parse_known_args for {args=}')

    util.debug(f'eda_config.get_eda_config: {parsed=} {unparsed=}  from {args=}')

    if parsed.config_yml:
        if not quiet:
            util.info(f'eda_config: --config-yml={parsed.config_yml} observed')
        fullpath = find_eda_config_yml_fpath(parsed.config_yml)
        config = get_config(fullpath)
        if not quiet:
            util.info(f'eda_config: using config: {fullpath}')

        # Calling get_config(fullpath) will add fullpath to config['config-yml'], so the
        # arg for --config-yml does not need to be re-added.
    else:
        config = None

    if parsed.config_yml != Defaults.config_yml:
        config = get_config_merged_with_defaults(config)

    return config, unparsed


def write_eda_config_and_args(
        dirpath : str, filename: str = EDA_OUTPUT_CONFIG_FNAME,
        command_obj_ref: object = None
) -> None:
    '''Writes and eda_config style dict to dirpath/filename'''
    if command_obj_ref is None:
        return
    fullpath = os.path.join(dirpath, filename)
    data = {}
    for x in ['command_name', 'config', 'target', 'args', 'modified_args', 'defines',
              'incdirs', 'files_v', 'files_sv', 'files_vhd', 'files_cpp', 'files_sdc',
              'files_non_source']:
        # Use deep copy b/c otherwise these are references to opencos.eda.
        data[x] = copy.deepcopy(getattr(command_obj_ref, x, ''))

    # copy util.args
    data['util'] = {
        'args': util.args
    }

    # fix some burried class references in command_obj_ref.config,
    # otherwise we won't be able to safe load this yaml, so cast as str repr.
    for k, v in getattr(command_obj_ref, 'config', {}).items():
        if k == 'command_handler':
            data['config'][k] = str(v)

    yaml_safe_writer(data=data, filepath=fullpath)



'''opencos.commands.deps_help - command handler for: eda deps_help [args]

Note this command is handled differently than others (such as CommandSim),
it is generally run as simply

    > eda deps_help
    > eda deps_help --verbose
    > eda deps_help --help

uses no tools and will print a help text regarding DEPS markup files to stdout.
'''


import os

from opencos.eda_base import Command
from opencos import util


BASIC_DEPS_HELP = '''

--------------------------------------------------------------------

  What is a DEPS.yml file and why does `eda` use this?
  - DEPS.yml is a fancy filelist.
  - Used to organize a project into "targets", a tool can run on a "target".
  - Allows for more than just source files attached to a "target".
    -- incdirs, defines, and args can be applied to a "target".

--------------------------------------------------------------------

  Hello World example:

  The following example is a DEPS.yml file example for a SystemVerilog simulation of
  hello_world_tb.sv. DEPS.yml is, in short, a fancy filelist. We use them in the `eda`
  app to organize projects.

--- DEPS.yml: ---

hello-world:           # <-- this is a named target that will be run

  deps:                # <-- 'deps' is a list of SV, Verilog, VHDL files in compile order
    - hello_world_tb.sv

  top: hello_world_tb  # <-- For testbenches, it is good practice to specifiy the topmost
                       #     module using using 'top'. This is not necessary for design
                       #     files.


--- hello_world_tb.sv: ---

module hello_world_tb;

  initial begin
    #10ns;
    $display("%t %m: Hello World!", $realtime);
    $display("%t %m: Test finished", $realtime);
    $finish;
  end

endmodule : hello_world_tb

---


  hello-world:
    The target name in the DEPS.yml we named is hello-world. That is a valid target
    that `eda` can use. Such as:

       eda sim --tool=verilator hello-world


--------------------------------------------------------------------

   Beyond Hello World example:

   The following example is a DEPS.yml file for a more complex module simulation.
   It has two files in ./DEPS.yml and ./lib/DEPS.yml.

--- ./DEPS.yml: ---

my_fifo:                         # <-- this is a design
  incdirs: . lib                 # <-- 'incdirs' define the paths searched to find `include files
  defines:
    FIFO_DEBUG                   # add a basic define
    FIFO_IMPLEMENTATION=uram     # add a define with a value
  deps:            # <-- 'deps' is a list of SV, Verilog, VHDL files in compile order
    - my_fifo.sv                 # an SV file pulled in directly
    - lib/bin_to_gray            # a target, in a subdirectory that has it's own DEPS

my_fifo_test:                    # <-- this is a TEST
  top: my_fifo_test              # the top will default to whatever target is provided
                                 # by the user, so this could be optional
  deps:
    - my_fifo                    # the target that is defined above
    - my_fifo_tb.sv              # an SV file pulled in directly

my_fifo_stress_test:             # <-- this is another TEST
  top: my_fifo_test              # not optional because top is not "my_fifo_stress_test"
  defines:
    STRESS_TEST                  # configures my_fifo_test to be more stressful
  deps:
    - my_fifo_test               # aside from the define, this is same as "my_fifo_test"

--- lib/DEPS.yml: ---

lib_pkg:                         # <-- this is a package required by bin_to_gray below
  deps:
    - assert_pkg.sv              # an SV package pulled in directly, before it's needed below
    - lib_pkg.sv                 # an SV package pulled in directly

bin_to_gray:                     # <-- this is the target that was required by ../my_fifo
  deps:
    - lib_pkg                    # a target package, listed first as SV requires packages
                                 # to be read before the code that uses them
    - bin_to_gray.sv             # an SV module pulled in directly

--------------------------------------------------------------------
'''


FULL_DEPS_HELP = '''

--------------------------------------------------------------------

   Full DEPS.yml schema:

```
DEFAULTS: # <table> defaults applied to ALL targets in this file, local targets ** override ** the defaults.

METADATA: # <table> unstructured data, any UPPERCASE first level key is not considered a target.

target-spec:

  args: # <array or | separated str>
    - --waves
    - --sim_plusargs="+info=500"

  defines: # <table>
    SOME_DEFINE: value
    SOME_DEFINE_NO_VALUE:   # we just leave this blank, or use nil (yaml's None)

  incdirs: # <array>
    - some/relative/path

  top: # <string>

  deps: # <array or | space separated string>
    - some_relative_target       # <string> aka, a target
    - some_file.sv               # <string> aka, a file
    - sv@some_file.txt           # <string> aka, ext@file where we'd like a file not ending in .sv to be
                                 # treated as a .sv file for tools.
                                 # Supported for sv@, v@, vhdl@, cpp@
    - commands:                  # <table> with key 'commands' for a <array>:  support for built-in commands
                                 # Note this cannot be confused for other targets or files.
      - shell: # <string>
        var-subst-args: # <bool> default false. If true, substitute vars in commands, such as {fpga}
                        # substituted from eda arg --fpga=SomeFpga, such that {fpga} becomes SomeFpga
        var-subst-os-env:  #<bool> default false. If true, substitute vars in commands using os.environ vars,
                           # such as {FPGA} could get substituted by env value for $FPGA
        tee: # <string> optional filename, otherwise shell commands write to {target-spec}__shell_0.log
        run-from-work-dir: #<bool> default true. If false, runs from the directory of this DEPS file.
        filepath-subst-target-dir: #<bool> default true. If false, disables shell file path
	                               substituion on this target's directory (this DEPS file dir).
        dirpath-subst-target-dir: #<bool> default false. If true, enables shell directory path
	                              substituion on this target's directory (this DEPS file dir).
      - shell: echo "Hello World!"
      - work-dir-add-sources: # <array or | space separated string>, this is how to add generated files
                              # to compile order list.
      - peakrdl:              # <string>     ## peakrdl command to generate CSRs

  reqs: # <array or | space separated string>
    - some_file.mem           # <string> aka, a non-source file required for this target.
                              # This file is checked for existence prior to invoking the tool involved, for example,
                              # in a simulation this would be done prior to a compile step.

  multi:
    ignore-this-target:  # <array of tables> eda commands to be ignored in `eda multi <command>` for this target only
                         # this is checked in the matching multi targets list, and is not inherited through dependencies.
      - commands: synth  # space separated strings
        tools: vivado    # space separated strings

      - commands: sim # omit tools, ignores 'sim' commands for all tools, for this target only, when this target
                      # is in the target list called by `eda multi`.

      - tools: vivado # omit commands, ignores all commands if tool is vivado, for this target only, when this target
                      # is in the target list called by `eda multi`.

    args: # <array> additional args added to all multi commands of this target.
          # Note that all args are POSIX with dashes, --sim-plusargs=value, etc.

  <eda-command>: # key is one of sim, flist, build, synth, etc.
                 # can be used instead of 'tags' to support different args or deps.
    disable-tools: # Note: not implemented yet.
    only-tools:    # Note: not implemented yet.
    args: # <array or | space separated string>
    deps: # <array or | space separated string> # Note: not implemented yet
    defines: ## <table>
    incdirs: ## <array>

  tags: # <table> this is the currently support tags features in a target.
    <tag-name>: # <string> key for table, can be anything, name is not used.
      with-tools: <array or | space separated string>
                  # If using one of these tools, apply these values.
                  # entries can be in the form: vivado, or vivado:2024.1
      with-commands: <array or | space separated string>
                  # apply if this was the `eda` command, such as: sim
      with-args: # <table> (optional) arg key/value pairs to match for this tag.
                 # this would be an alternative to running eda with --tags=value
                 # The existence of an argument with correct value would enable a tag.
                 # And example would be:
                 #   with-args:
                 #     waves: true
      args: <array or | space separated string> # args to be applied if this target is used, with a matching
                                          # tool in 'with-tools'.
      deps: <array or | space separated string, applied with tag>
      defines: <table, applied with tag>
      incdirs: <array, applied with tag>
      replace-config-tools: <table>  # spec matching eda_config_defaults.yml::tools.<tool> (replace merge strategy)
      additive-config-tools: <table> # spec matching eda_config_defaults.yml::tools.<tool> (additive merge strategy)


```
'''


class CommandDepsHelp:
    '''command handler for: eda deps-help'''

    command_name = 'deps-help'

    def __init__(self, config: dict):
        # We don't inherit opencos.eda_base.Command, so we have to set a few
        # member vars for Command.help to work.
        self.args = {}
        self.args_help = {}
        self.config = config
        self.status = 0

    def process_tokens( # pylint: disable=unused-argument
        self, tokens: list, process_all: bool = True,
        pwd: str = os.getcwd()
    ) -> list:
        '''This is effectively our 'run' method, entrypoint from opencos.eda.main'''

        print(BASIC_DEPS_HELP)
        if util.args['verbose'] or util.args['debug']:
            print()
            print(FULL_DEPS_HELP)

        return []

    def help(self, tokens: list) -> None:
        '''Since we don't inherit from opencos.eda_base.Command, need our own help
        method
        '''
        Command.help(self, tokens=tokens, no_targets=True)
        print()
        print(FULL_DEPS_HELP)



''' opencos.eda_tool_helper -- used by pytests and other checks to see if tools are loaded

which helps determine if a pytest is runnable for a given tool, or should be skipped.
Does this without calling `eda` or eda.main(..)

Example uses:
    from opencos import eda_tool_helper
    cfg, tools_loaded = eda_tool_helper.get_config_and_tools_loaded()
    assert 'verilator' in tools_loaded

'''


from opencos import eda, eda_config, util

# Used by pytest, so we can skip tests if tools aren't present.

def get_config_and_tools_loaded( # pylint: disable=dangerous-default-value
        quiet: bool = False, args: list = []
) -> (dict, set):
    '''Returns config dict and set tools_loaded, given the found config.

    Can BYO args such as --config-yml=MY_OWN_EDA_CONFIG.yml
    '''

    # We have to figure out what tools are avaiable w/out calling eda.main,
    # so we can get some of these using eda_config.get_eda_config()
    config, _ = eda_config.get_eda_config(args=args, quiet=quiet)
    config = eda.init_config(config=config, quiet=quiet)
    tools_loaded = config.get('tools_loaded', set()).copy()
    return config, tools_loaded


def get_all_handler_commands(config=None, tools_loaded=None) -> dict:
    '''Given a config and tools_loaded (or if not supplied uses defaults) returns a dict

    of { <command>: [list of tools that run that command, in auto-tool-order] }.

    For example:
       { "sim": ["verilator", "vivado"],
         "elab": ["slang", "verilator", ...], ...
       }
    '''
    all_handler_commands = {}

    if config is None or tools_loaded is None:
        config, tools_loaded = get_config_and_tools_loaded()

    assert isinstance(config, dict)
    assert isinstance(tools_loaded, set)

    # Let's re-walk auto_tools_order to get this ordered per eda command:
    for tool, table in config.get('auto_tools_order', [{}])[0].items():
        if tool not in tools_loaded:
            continue

        if table.get('disable-tools-multi', False):
            # Flagged as do-not-add when running eda command: tools-multi
            util.debug(f'eda_tool_helper.py -- skipping {tool=} it is set with flag',
                       'disable-tools-multi in config')
            continue

        for command in table.get('handlers', {}).keys():
            if command not in all_handler_commands:
                # create ordered list from config.
                all_handler_commands[command] = list([tool])
            else:
                all_handler_commands[command].append(tool)

    return all_handler_commands
